import json
import time
from typing import Any
from uuid import uuid4

import ray
from fastapi.openapi.utils import get_openapi
from fastapi.responses import StreamingResponse
from ray import serve

from aana.api.api_generation import Endpoint, add_custom_schemas_to_openapi_schema
from aana.api.app import app
from aana.api.event_handlers.event_manager import EventManager
from aana.api.responses import AanaJSONResponse
from aana.configs.settings import settings as aana_settings
from aana.core.models.chat import ChatCompletetion, ChatCompletionRequest, ChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.task import TaskId
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.storage.services.task import TaskInfo, delete_task, get_task_info


@serve.deployment(ray_actor_options={"num_cpus": 0.1})
@serve.ingress(app)
class RequestHandler:
    """This class is used to handle requests to the Aana application."""

    ready = False

    def __init__(self, endpoints: list[Endpoint]):
        """Constructor.

        Args:
            endpoints (dict): List of endpoints for the request
        """
        self.endpoints = endpoints

        self.event_manager = EventManager()
        self.custom_schemas: dict[str, dict] = {}
        for endpoint in self.endpoints:
            endpoint.register(
                app=app,
                custom_schemas=self.custom_schemas,
                event_manager=self.event_manager,
            )

        app.openapi = self.custom_openapi
        self.ready = True

    def custom_openapi(self) -> dict[str, Any]:
        """Returns OpenAPI schema, generating it if necessary."""
        if app.openapi_schema:
            return app.openapi_schema
        # TODO: populate title and version from package info
        openapi_schema = get_openapi(title="Aana", version="0.1.0", routes=app.routes)
        openapi_schema = add_custom_schemas_to_openapi_schema(
            openapi_schema=openapi_schema, custom_schemas=self.custom_schemas
        )
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    @app.get("/api/ready")
    async def is_ready(self):
        """The endpoint for checking if the application is ready.

        Real reason for this endpoint is to make automatic endpoint generation work.
        If RequestHandler doesn't have any endpoints defined manually,
        then the automatic endpoint generation doesn't work.
        #TODO: Find a better solution for this.

        Returns:
            AanaJSONResponse: The response containing the ready status.
        """
        return AanaJSONResponse(content={"ready": self.ready})

    async def call_endpoint(self, path: str, **kwargs: dict[str, Any]) -> Any:
        """Call the endpoint from FastAPI with the given name.

        Args:
            path (str): The path of the endpoint.
            **kwargs: The arguments to pass to the endpoint.

        Returns:
            Any: The response from the endpoint.
        """
        for e in self.endpoints:
            if e.path == path:
                endpoint = e
                break
        else:
            raise ValueError(f"Endpoint {path} not found")  # noqa: TRY003

        if not endpoint.initialized:
            await endpoint.initialize()

        if endpoint.is_streaming_response():
            return [item async for item in endpoint.run(**kwargs)]
        else:
            return await endpoint.run(**kwargs)

    @app.get(
        "/tasks/get/{task_id}",
        summary="Get Task Status",
        description="Get the task status by task ID.",
        include_in_schema=aana_settings.task_queue.enabled,
    )
    async def get_task_endpoint(self, task_id: str) -> TaskInfo:
        """Get the task with the given ID.

        Args:
            task_id (str): The ID of the task.

        Returns:
            TaskInfo: The status of the task.
        """
        return get_task_info(task_id)

    @app.get(
        "/tasks/delete/{task_id}",
        summary="Delete Task",
        description="Delete the task by task ID.",
        include_in_schema=aana_settings.task_queue.enabled,
    )
    async def delete_task_endpoint(self, task_id: str) -> TaskId:
        """Delete the task with the given ID.

        Args:
            task_id (str): The ID of the task.

        Returns:
            TaskInfo: The deleted task.
        """
        task = delete_task(task_id)
        return TaskId(task_id=str(task.id))

    @app.post("/chat/completions", response_model=ChatCompletetion)
    async def chat_completions(self, request: ChatCompletionRequest):
        """Handle chat completions requests for OpenAI compatible API."""

        async def _async_chat_completions(
            handle: AanaDeploymentHandle,
            dialog: ChatDialog,
            sampling_params: SamplingParams,
        ):
            async for response in handle.chat_stream(
                dialog=dialog, sampling_params=sampling_params
            ):
                chunk = {
                    "id": f"chatcmpl-{uuid4().hex}",
                    "object": "chat.completion.chunk",
                    "model": request.model,
                    "created": int(time.time()),
                    "choices": [
                        {
                            "index": 0,
                            "delta": {"content": response["text"], "role": "assistant"},
                        }
                    ],
                }
                yield f"data: {json.dumps(chunk)}\n\n"
            yield "data: [DONE]\n\n"

        # Check if the deployment exists
        try:
            handle = await AanaDeploymentHandle.create(request.model)
        except ray.serve.exceptions.RayServeException:
            return AanaJSONResponse(
                content={
                    "error": {"message": f"The model `{request.model}` does not exist."}
                },
                status_code=404,
            )

        # Check if the deployment is a chat model
        if not hasattr(handle, "chat") or not hasattr(handle, "chat_stream"):
            return AanaJSONResponse(
                content={
                    "error": {"message": f"The model `{request.model}` does not exist."}
                },
                status_code=404,
            )

        dialog = ChatDialog(
            messages=request.messages,
        )

        sampling_params = SamplingParams(
            temperature=request.temperature,
            max_tokens=request.max_tokens,
            top_p=request.top_p,
        )

        if request.stream:
            return StreamingResponse(
                _async_chat_completions(handle, dialog, sampling_params),
                media_type="application/x-ndjson",
            )
        else:
            response = await handle.chat(dialog=dialog, sampling_params=sampling_params)
            return {
                "id": f"chatcmpl-{uuid4().hex}",
                "object": "chat.completion",
                "model": request.model,
                "created": int(time.time()),
                "choices": [{"index": 0, "message": response["message"]}],
            }



from __future__ import annotations  # Let classes use themselves in type annotations

from typing import TYPE_CHECKING

from sqlalchemy import JSON, CheckConstraint, Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from aana.storage.models.base import BaseEntity, TimeStampEntity
from aana.storage.types import MediaIdSqlType

if TYPE_CHECKING:
    from aana.core.models.asr import (
        AsrSegments,
        AsrTranscription,
        AsrTranscriptionInfo,
    )
    from aana.core.models.media import MediaId


class TranscriptEntity(BaseEntity, TimeStampEntity):
    """ORM class for media transcripts generated by a model."""

    __tablename__ = "transcripts"

    id = Column(Integer, autoincrement=True, primary_key=True)
    model = Column(
        String, nullable=False, comment="Name of model used to generate transcript"
    )
    media_id = Column(
        MediaIdSqlType,
        ForeignKey("video.id"),
        nullable=False,
        comment="Foreign key to video table",
    )
    transcript = Column(String, comment="Full text transcript of media")
    segments = Column(JSON, comment="Segments of the transcript")
    language = Column(
        String, comment="Language of the transcript as predicted by model"
    )
    language_confidence = Column(
        Float,
        CheckConstraint(
            "0 <= language_confidence <= 1", name="language_confidence_value_range"
        ),
        comment="Confidence score of language prediction",
    )

    video = relationship("VideoEntity", back_populates="transcripts", uselist=False)

    @classmethod
    def from_asr_output(
        cls,
        model_name: str,
        media_id: MediaId,
        info: AsrTranscriptionInfo,
        transcription: AsrTranscription,
        segments: AsrSegments,
    ) -> TranscriptEntity:
        """Converts an AsrTranscriptionInfo and AsrTranscription to a single Transcript entity."""
        return TranscriptEntity(
            model=model_name,
            media_id=media_id,
            language=info.language,
            language_confidence=info.language_confidence,
            transcript=transcription.text,
            segments=[s.model_dump() for s in segments],
        )



import asyncio
import concurrent.futures
from typing import Any

import orjson
import ray
from pydantic import BaseModel, Field
from ray import serve

from aana.api.exception_handler import custom_exception_handler
from aana.configs.settings import settings as aana_settings
from aana.deployments.base_deployment import BaseDeployment
from aana.storage.models.task import Status as TaskStatus
from aana.storage.services.task import (
    get_task,
    get_unprocessed_tasks,
    update_task_status,
)
from aana.utils.asyncio import run_async


class TaskQueueConfig(BaseModel):
    """The configuration for the task queue deployment."""

    app_name: str = Field(description="The name of the Aana app")


@serve.deployment
class TaskQueueDeployment(BaseDeployment):
    """Deployment to serve the task queue."""

    def __init__(self):
        """Initialize the task queue deployment."""
        super().__init__()
        self.futures = {}
        loop = asyncio.get_running_loop()
        self.loop_task = loop.create_task(self.loop())
        self.loop_task.add_done_callback(
            lambda fut: fut.result() if not fut.cancelled() else None
        )

    def check_health(self):
        """Check the health of the deployment."""
        # if the loop is not running, the deployment is unhealthy
        if self.loop_task.done():
            raise RuntimeError("Task queue loop is not running")  # noqa: TRY003

    def __del__(self):
        """Clean up the deployment."""
        # Cancel the loop task to prevent tasks from being reassigned
        self.loop_task.cancel()
        # Set all non-completed tasks to NOT_FINISHED
        for task_id, future in self.futures.items():
            if not future.done():
                update_task_status(task_id, TaskStatus.NOT_FINISHED, 0)

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        The configuration should conform to the TaskQueueConfig schema.
        """
        config_obj = TaskQueueConfig(**config)
        self.handle = None
        self.app_name = config_obj.app_name
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=aana_settings.task_queue.num_workers,
        )

    async def loop(self):  # noqa: C901
        """The main loop for the task queue deployment.

        The loop will check the queue and assign tasks to workers.
        """

        async def handle_task(task_id: str):
            """Process a task."""
            # Fetch the task details
            task = get_task(task_id)
            # Initially set the task status to RUNNING
            update_task_status(task_id, TaskStatus.RUNNING, 0)
            try:
                # Call the endpoint asynchronously
                out = await self.handle.call_endpoint.remote(task.endpoint, **task.data)
                # Update the task status to COMPLETED
                update_task_status(task_id, TaskStatus.COMPLETED, 100, out)
            except Exception as e:
                # Handle the exception and update the task status to FAILED
                error_response = custom_exception_handler(None, e)
                error = orjson.loads(error_response.body)
                update_task_status(task_id, TaskStatus.FAILED, 0, error)

        def run_handle_task(task_id):
            """Wrapper to run the handle_task function."""
            run_async(handle_task(task_id))

        def is_thread_pool_full():
            """Check if the thread pool has too many tasks.

            We use it to stop assigning tasks to the thread pool if it's full
            to prevent the thread pool from being overwhelmed.
            We don't want to schedule all tasks from the task queue (could be millions).
            """
            return (
                self.thread_pool._work_queue.qsize()
                > aana_settings.task_queue.num_workers * 2
            )

        while True:
            if not self.configured:
                # Wait for the deployment to be configured.
                await asyncio.sleep(1)
                continue

            # Remove completed tasks from the futures dictionary
            for task_id in list(self.futures.keys()):
                if self.futures[task_id].done():
                    del self.futures[task_id]

            if is_thread_pool_full():
                # wait a bit to give the thread pool time to process tasks
                await asyncio.sleep(0.1)
                continue

            tasks = get_unprocessed_tasks(
                limit=aana_settings.task_queue.num_workers * 2
            )

            if not tasks:
                await asyncio.sleep(0.1)
                continue

            if not self.handle:
                # Sometimes the app isn't available immediately after the deployment is created
                # so we need to wait for it to become available
                for _ in range(10):
                    try:
                        self.handle = serve.get_app_handle(self.app_name)
                        break
                    except ray.serve.exceptions.RayServeException as e:
                        print(
                            f"App {self.app_name} not available yet: {e}, retrying..."
                        )
                        await asyncio.sleep(1)
                else:
                    # If the app is not available after all retries, try again
                    # but without catching the exception
                    # (if it fails, the deployment will be unhealthy, and restart will be attempted)
                    self.handle = serve.get_app_handle(self.app_name)

            for task in tasks:
                if is_thread_pool_full():
                    # wait a bit to give the thread pool time to process tasks
                    await asyncio.sleep(0.1)
                    break
                update_task_status(task.id, TaskStatus.ASSIGNED, 0)
                future = self.thread_pool.submit(run_handle_task, task.id)
                self.futures[task.id] = future



from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.exceptions.db import MediaIdAlreadyExistsException, NotFoundException
from aana.storage.models.media import MediaEntity
from aana.storage.repository.base import BaseRepository


class MediaRepository(BaseRepository[MediaEntity]):
    """Repository for media files."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, MediaEntity)

    def check_media_exists(self, media_id: MediaId) -> bool:
        """Checks if a media file exists in the database.

        Args:
            media_id (MediaId): The media ID.

        Returns:
            bool: True if the media exists, False otherwise.
        """
        try:
            self.read(media_id)
        except NotFoundException:
            return False

        return True

    def create(self, entity: MediaEntity) -> MediaEntity:
        """Inserts a single new entity.

        Args:
            entity (MediaEntity): The entity to insert.

        Returns:
            MediaEntity: The inserted entity.
        """
        if self.check_media_exists(entity.id):
            raise MediaIdAlreadyExistsException(self.table_name, entity.id)

        return super().create(entity)



from ray import serve

from aana.utils.typing import is_async_generator


class AanaDeploymentHandle:
    """A handle to interact with a deployed Aana deployment.

    Use create method to create a deployment handle.

    ```python
    deployment_handle = await AanaDeploymentHandle.create("deployment_name")
    ```

    Attributes:
        deployment_name (str): The name of the deployment.
    """

    def __init__(self, deployment_name: str):
        """A handle to interact with a deployed Aana deployment.

        Args:
            deployment_name (str): The name of the deployment.
        """
        self.handle = serve.get_app_handle(deployment_name)
        self.deployment_name = deployment_name
        self.methods = None

    def create_async_method(self, name: str):
        """Create an method to interact with the deployment.

        Args:
            name (str): The name of the method.
        """
        method_info = self.methods[name]
        annotations = method_info.get("annotations", {})
        return_type = annotations.get("return", None)

        if is_async_generator(return_type):

            async def method(*args, **kwargs):
                async for item in self.handle.options(
                    method_name=name, stream=True
                ).remote(*args, **kwargs):
                    yield item
        else:

            async def method(*args, **kwargs):
                return await self.handle.options(method_name=name).remote(
                    *args, **kwargs
                )

        if "annotations" in self.methods[name]:
            method.__annotations__ = self.methods[name]["annotations"]
        if "doc" in self.methods[name]:
            method.__doc__ = self.methods[name]["doc"]
        return method

    async def load_methods(self):
        """Load the methods available in the deployment."""
        self.methods = await self.handle.get_methods.remote()
        for name in self.methods:
            setattr(self, name, self.create_async_method(name))

    @classmethod
    async def create(cls, deployment_name: str):
        """Create a deployment handle.

        Args:
            deployment_name (str): The name of the deployment to interact with.
        """
        handle = cls(deployment_name)
        await handle.load_methods()
        return handle



from aana.api.event_handlers.rate_limit_handler import RateLimitHandler
from aana.configs.deployments import (
    stablediffusion2_deployment,
)
from aana.projects.stablediffusion2.endpoints import ImageGenerationEndpoint
from aana.sdk import AanaSDK

deployments = [
    {
        "name": "image_generation_deployment",
        "instance": stablediffusion2_deployment,
    },
]

endpoints = [
    {
        "name": "generate_image",
        "path": "/generate_image",
        "summary": "Generates an image from a text prompt",
        "endpoint_cls": ImageGenerationEndpoint,
    },
    {
        "name": "generate_image_rate_limited",
        "path": "/generate_image_rate_limited",
        "summary": "Generates an image from a text prompt",
        "endpoint_cls": ImageGenerationEndpoint,
        "event_handlers": [RateLimitHandler(1, 30)],
    },
]

aana_app = AanaSDK(name="stablediffusion2")

for deployment in deployments:
    aana_app.register_deployment(
        name=deployment["name"],
        instance=deployment["instance"],
    )

for endpoint in endpoints:
    aana_app.register_endpoint(
        name=endpoint["name"],
        path=endpoint["path"],
        summary=endpoint["summary"],
        endpoint_cls=endpoint["endpoint_cls"],
        event_handlers=endpoint.get("event_handlers", []),
    )



from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.storage.models.caption import CaptionEntity
from aana.storage.repository.base import BaseRepository


class CaptionRepository(BaseRepository[CaptionEntity]):
    """Repository for Captions."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, CaptionEntity)

    def get_captions(self, model_name: str, media_id: MediaId) -> list[CaptionEntity]:
        """Get the captions for a video.

        Args:
            model_name (str): The model name.
            media_id (MediaId): The media ID.

        Returns:
            list[CaptionEntity]: The list of caption entities.
        """
        entities: list[CaptionEntity] = (
            self.session.query(self.model_class)
            .filter_by(media_id=media_id, model=model_name)
            .order_by(self.model_class.frame_id)
            .all()
        )
        return entities



from collections.abc import AsyncGenerator
from typing import Annotated, TypedDict

from pydantic import Field

from aana.api.api_generation import Endpoint
from aana.core.models.chat import ChatDialog, Prompt
from aana.core.models.sampling import SamplingParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle


class LlmGenerateEndpointOutput(TypedDict):
    """LLM generate endpoint output."""

    completion: Annotated[str, Field(description="The generated text.")]


class LlmGenerateEndpoint(Endpoint):
    """LLM generate endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, prompt: Prompt, sampling_params: SamplingParams
    ) -> LlmGenerateEndpointOutput:
        """Run the LLM generate endpoint."""
        llm_output = await self.llm_handle.generate(
            prompt=prompt, sampling_params=sampling_params
        )
        return LlmGenerateEndpointOutput(completion=llm_output["text"])


class LlmGenerateStreamEndpoint(Endpoint):
    """LLM generate stream endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, prompt: Prompt, sampling_params: SamplingParams
    ) -> AsyncGenerator[LlmGenerateEndpointOutput, None]:
        """Run the LLM generate stream endpoint."""
        async for item in self.llm_handle.generate_stream(
            prompt=prompt, sampling_params=sampling_params
        ):
            yield LlmGenerateEndpointOutput(completion=item["text"])


class LlmChatEndpointOutput(TypedDict):
    """LLM chat endpoint output."""

    message: Annotated[str, Field(description="The response message.")]


class LlmChatEndpoint(Endpoint):
    """LLM chat endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, dialog: ChatDialog, sampling_params: SamplingParams
    ) -> LlmChatEndpointOutput:
        """Run the LLM chat endpoint."""
        llm_output = await self.llm_handle.chat(
            dialog=dialog, sampling_params=sampling_params
        )
        return LlmChatEndpointOutput(message=llm_output["message"])


class LlmChatStreamEndpointOutput(TypedDict):
    """LLM chat stream endpoint output."""

    completion: Annotated[str, Field(description="Chunk of generated text.")]


class LlmChatStreamEndpoint(Endpoint):
    """LLM chat stream endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, dialog: ChatDialog, sampling_params: SamplingParams
    ) -> AsyncGenerator[LlmChatStreamEndpointOutput, None]:
        """Run the LLM chat stream endpoint."""
        async for item in self.llm_handle.chat_stream(
            dialog=dialog, sampling_params=sampling_params
        ):
            yield LlmChatStreamEndpointOutput(completion=item["text"])



from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.exceptions.db import NotFoundException
from aana.storage.models.transcript import TranscriptEntity
from aana.storage.repository.base import BaseRepository


class TranscriptRepository(BaseRepository[TranscriptEntity]):
    """Repository for Transcripts."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session=session, model_class=TranscriptEntity)

    def get_transcript(self, model_name: str, media_id: MediaId) -> TranscriptEntity:
        """Get the transcript for a video.

        Args:
            model_name (str): The name of the model used to generate the transcript.
            media_id (MediaId): The media ID.

        Returns:
            TranscriptEntity: The transcript entity.

        Raises:
            NotFoundException: The transcript does not exist.
        """
        entity: TranscriptEntity | None = (
            self.session.query(self.model_class)
            .filter_by(model=model_name, media_id=media_id)
            .first()
        )
        if not entity:
            raise NotFoundException(self.table_name, media_id)
        return entity



# ruff: noqa: S101
import json
from importlib import resources
from pathlib import Path

import pytest
from ray import serve

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_to_dict
from aana.core.models.vad import VadParams
from aana.tests.utils import (
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)


def compare_vad_outputs(expected_output, predictions):
    """Compare two vad outputs.

    Start and end positions are compared using Levenshtein distance.

    Args:
        expected_output (dict(list[dict])): the expected vad_output
        predictions (dict(list[dict])): the predictions of vad

    Raises:
        AssertionError: if vad_outputs differ too much
    """
    # Number of 30 sec segments should not change.
    assert len(expected_output["segments"]) == len(predictions["segments"])

    if len(expected_output["segments"]) != 0:  # non-empty files
        # all start and end times should be similar and number of segments should be same.
        for expected, predicted in zip(
            expected_output["segments"], predictions["segments"], strict=False
        ):
            assert (
                abs(
                    expected["time_interval"]["start"]
                    - predicted["time_interval"]["start"]
                )
                < 2.0
            )  # all segment starts within 2 sec
            assert (
                abs(
                    expected["time_interval"]["end"] - predicted["time_interval"]["end"]
                )
                < 2.0
            )  # all segment ends within 2 sec

            # check same number of small voiced segments within each vad segment
            assert len(expected["segments"]) == len(predicted["segments"])


@pytest.fixture(scope="function", params=get_deployments_by_type("VadDeployment"))
def setup_vad_deployment(app_setup, request):
    """Setup vad deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "vad_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
@pytest.mark.parametrize("audio_file", ["physicsworks.wav", "squirrel.wav"])
async def test_vad_deployment(setup_vad_deployment, audio_file):
    """Test vad deployment."""
    handle = serve.get_app_handle("vad_deployment")

    audio_file_name = Path(audio_file).stem
    expected_output_path = resources.path(
        "aana.tests.files.expected.vad", f"{audio_file_name}_vad.json"
    )
    assert (
        expected_output_path.exists()
    ), f"Expected output not found: {expected_output_path}"
    with Path(expected_output_path) as path, path.open() as f:
        expected_output = json.load(f)

    # asr_preprocess_vad method to test
    path = resources.path("aana.tests.files.audios", audio_file)
    assert path.exists(), f"Audio not found: {path}"

    audio = Audio(path=path)

    output = await handle.asr_preprocess_vad.remote(audio=audio, params=VadParams())
    output = pydantic_to_dict(output)
    compare_vad_outputs(expected_output, output)



# This file is used to define fixtures that are used in the integration tests.
# The fixtures are used to setup Ray and Ray Serve, and to call the endpoints.
# The fixtures depend on each other, to setup the environment for the tests.
# Here is a dependency graph of the fixtures:
# app_setup (module scope, starts Ray and Ray Serve app for a specific target, args: deployments, endpoints)
#     -> call_endpoint (module scope, calls endpoint, args: endpoint_path, data, ignore_expected_output, expected_error)

# ruff: noqa: S101
import importlib
import os
import tempfile
from pathlib import Path

import pytest

from aana.configs.db import DbSettings, SQLiteConfig
from aana.configs.settings import settings as aana_settings
from aana.sdk import AanaSDK
from aana.storage.op import DbType
from aana.tests.utils import (
    call_and_check_endpoint,
    clear_database,
    is_gpu_available,
    is_using_deployment_cache,
)
from aana.utils.json import jsonify


@pytest.fixture(scope="module")
def app_setup():
    """Setup Ray Serve app for given deployments and endpoints."""
    # create temporary database
    tmp_database_path = Path(tempfile.mkstemp(suffix=".db")[1])
    db_config = DbSettings(
        datastore_type=DbType.SQLITE,
        datastore_config=SQLiteConfig(path=tmp_database_path),
    )
    # set environment variable for the database config so Ray can find it
    os.environ["DB_CONFIG"] = jsonify(db_config)

    # set database config in aana settings
    aana_settings.db_config = db_config

    # clear database before running the test
    clear_database(aana_settings)

    app = AanaSDK()
    app.connect(
        port=8000, show_logs=True, num_cpus=10
    )  # pretend we have 10 cpus for testing

    def start_app(deployments, endpoints):
        for deployment in deployments:
            deployment_instance = deployment["instance"]
            if not is_gpu_available() and is_using_deployment_cache():
                # if GPU is not available and we are using deployment cache,
                # then we don't want to request GPU resources
                deployment_instance = deployment_instance.options(
                    ray_actor_options={"num_gpus": 0}
                )

            app.register_deployment(
                name=deployment["name"], instance=deployment_instance
            )

        for endpoint in endpoints:
            app.register_endpoint(
                name=endpoint["name"],
                path=endpoint["path"],
                summary=endpoint["summary"],
                endpoint_cls=endpoint["endpoint_cls"],
                event_handlers=endpoint.get("event_handlers", []),
            )

        app.deploy(blocking=False)

        return app

    yield start_app

    # delete temporary database
    tmp_database_path.unlink()

    app.shutdown()


@pytest.fixture(scope="module")
def call_endpoint(app_setup, request):  # noqa: D417
    """Call endpoint.

    Args:
        endpoint_path: The endpoint path.
        data: The data to send.
        ignore_expected_output: Whether to ignore the expected output. Defaults to False.
        expected_error: The expected error. Defaults to None.
    """
    target = request.param

    module = importlib.import_module(f"aana.projects.{target}.app")
    deployments = module.deployments
    endpoints = module.endpoints

    aana_app = app_setup(deployments, endpoints)

    port = aana_app.port
    route_prefix = ""

    def _call_endpoint(
        endpoint_path: str,
        data: dict,
        ignore_expected_output: bool = False,
        expected_error: str | None = None,
    ) -> dict | list:
        endpoint = None
        for e in endpoints:
            if e["path"] == endpoint_path:
                endpoint = e
                break
        if endpoint is None:
            raise ValueError(f"Endpoint with path {endpoint_path} not found")  # noqa: TRY003
        is_streaming = endpoint["endpoint_cls"].is_streaming_response()

        return call_and_check_endpoint(
            target=target,
            port=port,
            route_prefix=route_prefix,
            endpoint_path=endpoint_path,
            data=data,
            is_streaming=is_streaming,
            expected_error=expected_error,
            ignore_expected_output=ignore_expected_output,
        )

    return _call_endpoint


@pytest.fixture(scope="module")
def one_request_worker():
    """Fixture to update settings to only run one request worker."""
    aana_settings.num_workers = 1
    yield



# ruff: noqa: S101
import haystack
import haystack.components.preprocessors
import PIL
import pytest

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.integrations.haystack.deployment_component import AanaDeploymentComponent
from aana.tests.deployments.test_stablediffusion2_deployment import (
    setup_deployment as setup_stablediffusion2_deployment,  # noqa: F401
)
from aana.tests.utils import is_gpu_available, is_using_deployment_cache


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
async def test_haystack_wrapper(setup_stablediffusion2_deployment):  # noqa: F811
    """Tests haystack wrapper for deployments."""
    deployment_name = "sd2_deployment"
    method_name = "generate"
    result_key = "image"
    deployment_handle = await AanaDeploymentHandle.create(deployment_name)
    component = AanaDeploymentComponent(deployment_handle, method_name)
    result = component.run(prompt="foo")
    assert result_key in result, result


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
async def test_haystack_pipeline(setup_stablediffusion2_deployment):  # noqa: F811
    """Tests haystack wrapper in a pipeline."""

    # Haystack components generally take lists of things
    # so we need a component to turn a list of strings into a single string
    @haystack.component
    class TextCombinerComponent:
        @haystack.component.output_types(text=str)
        def run(self, texts: list[str]):
            return {"text": " ".join(texts)}

    # Haystack also doesn't have any simple components that
    # take images as inputs, so let's have one of those as well
    @haystack.component
    class ImageSizeComponent:
        @haystack.component.output_types(size=tuple[int, int])
        def run(self, image: PIL.Image.Image):
            return {"size": image.size}

    deployment_name = "sd2_deployment"
    method_name = "generate"
    # result_key = "image"
    deployment_handle = await AanaDeploymentHandle.create(deployment_name)
    aana_component = AanaDeploymentComponent(deployment_handle, method_name)
    aana_component.warm_up()
    text_cleaner = haystack.components.preprocessors.TextCleaner(
        convert_to_lowercase=False, remove_punctuation=True, remove_numbers=True
    )
    text_combiner = TextCombinerComponent()
    image_sizer = ImageSizeComponent()

    pipeline = haystack.Pipeline()
    pipeline.add_component("stablediffusion2", aana_component)
    pipeline.add_component("text_cleaner", text_cleaner)
    pipeline.add_component("text_combiner", text_combiner)
    pipeline.add_component("image_sizer", image_sizer)
    pipeline.connect("text_cleaner.texts", "text_combiner.texts")
    pipeline.connect("text_combiner.text", "stablediffusion2.prompt")
    pipeline.connect("stablediffusion2.image", "image_sizer.image")
    result = pipeline.run({"text_cleaner": {"texts": ["A? dog!"]}})

    assert "image_sizer" in result
    assert "size" in result["image_sizer"]
    assert len(result["image_sizer"]["size"]) == 2


@pytest.mark.asyncio
async def test_haystack_wrapper_fails(setup_stablediffusion2_deployment):  # noqa: F811
    """Tests that haystack wrapper raises if method_name is missing."""
    deployment_name = "sd2_deployment"
    missing_method_name = "does_not_exist"
    deployment_handle = await AanaDeploymentHandle.create(deployment_name)
    with pytest.raises(AttributeError):
        _component = AanaDeploymentComponent(deployment_handle, missing_method_name)
